# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest import TestCase

from flexmock import flexmock, flexmock_teardown
from hamcrest import assert_that, is_

from netman.core.objects.exceptions import NetmanException
from netman.core.objects.flow_control_switch import FlowControlSwitch
from netman.core.objects.switch_base import SwitchBase
from netman.core.objects.switch_descriptor import SwitchDescriptor


class FlowControlSwitchTest(TestCase):

    def setUp(self):
        self.wrapped_switch = flexmock(SwitchBase(SwitchDescriptor("cisco", "name")))
        self.lock = flexmock()
        self.switch = FlowControlSwitch(self.wrapped_switch, self.lock)

    def tearDown(self):
        flexmock_teardown()

    def test_a_get_method_connects_and_executes(self):
        self.wrapped_switch.should_receive("_connect").once().ordered()
        self.wrapped_switch.should_receive("get_vlan").once().ordered().with_args(1000)
        self.wrapped_switch.should_receive("_disconnect").once().ordered()

        self.switch.get_vlan(1000)

    def test_a_get_method_still_dc_if_raising(self):
        self.wrapped_switch.should_receive("_connect").once().ordered()
        self.wrapped_switch.should_receive("get_vlan").once().ordered().with_args(1000).and_raise(NetmanException)
        self.wrapped_switch.should_receive("_disconnect").once().ordered()

        with self.assertRaises(NetmanException):
            self.switch.get_vlan(1000)

    def test_a_get_method_connect_fail_gives_up(self):
        self.wrapped_switch.should_receive("_connect").once().ordered().and_raise(NetmanException)

        with self.assertRaises(NetmanException):
            self.switch.get_vlan(1000)

    def test_a_get_method_does_not_connect_if_already_connected(self):
        self.wrapped_switch.connected = True

        self.wrapped_switch.should_receive("get_vlan").once().ordered().with_args(1000)

        self.switch.get_vlan(1000)

    def test_an_operation_method_locks_connects_and_executes_in_a_transaction(self):

        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered()
        self.wrapped_switch.should_receive("add_vlan").once().ordered().with_args(1000)
        self.wrapped_switch.should_receive("commit_transaction").once().ordered()
        self.wrapped_switch.should_receive("_end_transaction").once().ordered()
        self.wrapped_switch.should_receive("_disconnect").once().ordered()
        self.lock.should_receive("release").once().ordered()

        self.switch.add_vlan(1000)

    def test_an_operation_method_unlocks_and_rollback_if_raising(self):

        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered()
        self.wrapped_switch.should_receive("add_vlan").once().ordered().with_args(1000).and_raise(NetmanException)
        self.wrapped_switch.should_receive("rollback_transaction").once().ordered()
        self.wrapped_switch.should_receive("_end_transaction").once().ordered()
        self.wrapped_switch.should_receive("_disconnect").once().ordered()
        self.lock.should_receive("release").once().ordered()

        with self.assertRaises(NetmanException):
            self.switch.add_vlan(1000)

    def test_an_operation_method_failing_to_start_transaction_rollbacks_unlocks_dc(self):

        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered().and_raise(NetmanException)
        self.wrapped_switch.should_receive("_disconnect").once().ordered()
        self.lock.should_receive("release").once().ordered()

        with self.assertRaises(NetmanException):
            self.switch.add_vlan(1000)

    def test_an_operation_method_failing_to_end_transaction_unlocks_dc(self):

        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered()
        self.wrapped_switch.should_receive("add_vlan").once().ordered().with_args(1000)
        self.wrapped_switch.should_receive("commit_transaction").once().ordered()
        self.wrapped_switch.should_receive("_end_transaction").once().ordered().and_raise(NetmanException)
        self.wrapped_switch.should_receive("_disconnect").once().ordered()
        self.lock.should_receive("release").once().ordered()

        with self.assertRaises(NetmanException):
            self.switch.add_vlan(1000)

    def test_an_operation_method_failing_to_commit_transaction_unlocks_dc(self):

        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered()
        self.wrapped_switch.should_receive("add_vlan").once().ordered().with_args(1000)
        self.wrapped_switch.should_receive("commit_transaction").once().ordered().and_raise(NetmanException)
        self.wrapped_switch.should_receive("rollback_transaction").once().ordered()
        self.wrapped_switch.should_receive("_end_transaction").once().ordered()
        self.wrapped_switch.should_receive("_disconnect").once().ordered()
        self.lock.should_receive("release").once().ordered()

        with self.assertRaises(NetmanException):
            self.switch.add_vlan(1000)

    def test_an_operation_method_failing_to_disconnect_unlocks(self):

        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered()
        self.wrapped_switch.should_receive("add_vlan").once().ordered().with_args(1000)
        self.wrapped_switch.should_receive("commit_transaction").once().ordered()
        self.wrapped_switch.should_receive("_end_transaction").once().ordered()
        self.wrapped_switch.should_receive("_disconnect").once().ordered().and_raise(NetmanException)
        self.lock.should_receive("release").once().ordered()

        with self.assertRaises(NetmanException):
            self.switch.add_vlan(1000)

    def test_an_operation_method_failing_to_connect_unlocks(self):

        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").once().ordered().and_raise(NetmanException)
        self.lock.should_receive("release").once().ordered()

        with self.assertRaises(NetmanException):
            self.switch.add_vlan(1000)

    def test_an_operation_method_failing_to_lock_gives_up(self):

        self.lock.should_receive("acquire").once().ordered().and_raise(NetmanException)

        with self.assertRaises(NetmanException):
            self.switch.add_vlan(1000)

    def test_an_operation_method_already_connected_locks_and_executes_in_a_transaction(self):
        self.wrapped_switch.connected = True

        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").never()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered()
        self.wrapped_switch.should_receive("add_vlan").once().ordered().with_args(1000)
        self.wrapped_switch.should_receive("commit_transaction").once().ordered()
        self.wrapped_switch.should_receive("_end_transaction").once().ordered()
        self.wrapped_switch.should_receive("_disconnect").never()
        self.lock.should_receive("release").once().ordered()

        self.switch.add_vlan(1000)

    def test_an_operation_method_already_connected_and_in_transaction_just_executes(self):
        self.wrapped_switch.connected = True
        self.wrapped_switch.in_transaction = True

        self.wrapped_switch.should_receive("add_vlan").once().ordered().with_args(1000)

        self.switch.add_vlan(1000)

    def test_transaction_context_locks_connect_and_execute(self):

        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered()
        self.wrapped_switch.should_receive("add_vlan").once().ordered().with_args(1000)
        self.wrapped_switch.should_receive("commit_transaction").once().ordered()
        self.wrapped_switch.should_receive("_end_transaction").once().ordered()
        self.wrapped_switch.should_receive("_disconnect").once().ordered()
        self.lock.should_receive("release").once().ordered()

        with self.switch.transaction():
            self.switch.add_vlan(1000)

    def test_transaction_context_unlocks_rollback_and_dc_if_fails(self):

        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered()
        self.wrapped_switch.should_receive("add_vlan").once().ordered().with_args(1000).and_raise(NetmanException)
        self.wrapped_switch.should_receive("rollback_transaction").once().ordered()
        self.wrapped_switch.should_receive("_end_transaction").once().ordered()
        self.wrapped_switch.should_receive("_disconnect").once().ordered()
        self.lock.should_receive("release").once().ordered()

        with self.assertRaises(NetmanException):
            with self.switch.transaction():
                self.switch.add_vlan(1000)

    def test_transaction_context_already_connected_only_locks_connect_and_execute(self):
        self.wrapped_switch.connected = True

        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered()
        self.wrapped_switch.should_receive("add_vlan").once().ordered().with_args(1000)
        self.wrapped_switch.should_receive("commit_transaction").once().ordered()
        self.wrapped_switch.should_receive("_end_transaction").once().ordered()
        self.lock.should_receive("release").once().ordered()

        with self.switch.transaction():
            self.switch.add_vlan(1000)

    def test_transaction_context_already_in_transaction_only_executes(self):
        self.wrapped_switch.connected = True
        self.wrapped_switch.in_transaction = True

        self.wrapped_switch.should_receive("add_vlan").once().ordered().with_args(1000)

        with self.switch.transaction():
            self.switch.add_vlan(1000)

    def test_switch_contract_compliance_start_transaction(self):

        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered()

        self.switch.start_transaction()

    def test_switch_contract_compliance_start_transaction_failing_to_start_rollback_dc_unlocks(self):

        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered().and_raise(NetmanException)
        self.wrapped_switch.should_receive("_disconnect").once().ordered()
        self.lock.should_receive("release").once().ordered()

        with self.assertRaises(NetmanException):
            self.switch.start_transaction()

    def test_switch_contract_compliance_start_transaction_failing_to_connect_unlocks(self):

        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").once().ordered().and_raise(NetmanException)
        self.lock.should_receive("release").once().ordered()

        with self.assertRaises(NetmanException):
            self.switch.start_transaction()

    def test_switch_contract_compliance_start_transaction_failing_to_lock_gives_up(self):

        self.lock.should_receive("acquire").once().ordered().and_raise(NetmanException)

        with self.assertRaises(NetmanException):
            self.switch.start_transaction()

    def test_switch_contract_compliance_start_transaction_already_connected_don_t_connect(self):
        self.wrapped_switch.connected = True

        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered()

        self.switch.start_transaction()

    def test_switch_contract_compliance_start_transaction_already_in_transaction_does_nothing(self):
        self.wrapped_switch.connected = True
        self.wrapped_switch.in_transaction = True

        self.switch.start_transaction()

    def test_switch_contract_compliance_end_transaction(self):
        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered()

        self.switch.start_transaction()

        self.wrapped_switch.should_receive("_end_transaction").once().ordered()
        self.wrapped_switch.should_receive("_disconnect").once().ordered()
        self.lock.should_receive("release").once().ordered()

        self.switch.end_transaction()

    def test_switch_contract_compliance_end_transaction_dont_dc_for_a_legit_connection_after_a_managed_one(self):
        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered()

        self.switch.start_transaction()

        self.wrapped_switch.should_receive("_end_transaction").once().ordered()
        self.wrapped_switch.should_receive("_disconnect").once().ordered()
        self.lock.should_receive("release").once().ordered()

        self.switch.end_transaction()

        self.wrapped_switch.connected = True

        self.wrapped_switch.should_receive("_end_transaction").once().ordered()
        self.lock.should_receive("release").once().ordered()

        self.switch.end_transaction()

    def test_switch_contract_compliance_end_transaction_if_not_connected_by_start_dont_dc(self):
        self.wrapped_switch.connected = True

        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered()

        self.switch.start_transaction()

        self.wrapped_switch.should_receive("_end_transaction").once().ordered()
        self.lock.should_receive("release").once().ordered()

        self.switch.end_transaction()

    def test_switch_contract_compliance_end_transaction_failing_end_unlocks_and_dc(self):
        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered()

        self.switch.start_transaction()

        self.wrapped_switch.should_receive("_end_transaction").once().ordered().and_raise(NetmanException)
        self.wrapped_switch.should_receive("_disconnect").once().ordered()
        self.lock.should_receive("release").once().ordered()

        with self.assertRaises(NetmanException):
            self.switch.end_transaction()

    def test_switch_contract_compliance_end_transaction_failing_dc_unlocks(self):
        self.lock.should_receive("acquire").once().ordered()
        self.wrapped_switch.should_receive("_connect").once().ordered()
        self.wrapped_switch.should_receive("_start_transaction").once().ordered()

        self.switch.start_transaction()

        self.wrapped_switch.should_receive("_end_transaction").once().ordered()
        self.wrapped_switch.should_receive("_disconnect").once().ordered().and_raise(NetmanException)
        self.lock.should_receive("release").once().ordered()

        with self.assertRaises(NetmanException):
            self.switch.end_transaction()

    def test_switch_contract_compliance_connect(self):
        self.wrapped_switch.should_receive("connect").once().ordered()
        self.switch.connect()

    def test_switch_contract_compliance_disconnect(self):
        self.wrapped_switch.should_receive("disconnect").once().ordered()
        self.switch.disconnect()

    def test_switch_contract_compliance_commit_transaction(self):
        self.wrapped_switch.should_receive("commit_transaction").once().ordered()
        self.switch.commit_transaction()

    def test_switch_contract_compliance_rollback_transaction(self):
        self.wrapped_switch.should_receive("rollback_transaction").once().ordered()
        self.switch.rollback_transaction()

    def test_switch_contract_compliance_switch_descriptor(self):
        assert_that(self.switch.switch_descriptor, is_(self.wrapped_switch.switch_descriptor))



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import textwrap
import unittest

from flexmock import flexmock, flexmock_teardown
from hamcrest import assert_that, equal_to, instance_of, contains_string, has_length, is_
from ncclient.operations import RPCError
from ncclient.xml_ import to_ele
from netaddr import IPAddress, IPNetwork
from netman.adapters.switches.juniper.base import Juniper
from netman.adapters.switches.juniper.mx import netconf
from netman.core.objects.access_groups import IN, OUT
from netman.core.objects.exceptions import VlanAlreadyExist, BadVlanNumber, BadVlanName, UnknownVlan, \
    IPAlreadySet, UnknownIP, InterfaceInWrongPortMode, AccessVlanNotSet, UnknownInterface, TrunkVlanNotSet, \
    VrrpDoesNotExistForVlan
from netman.core.objects.port_modes import ACCESS
from netman.core.objects.switch_descriptor import SwitchDescriptor
from netman.core.switch_factory import RealSwitchFactory
from tests.adapters.switches.juniper_test import an_ok_response, is_xml, a_configuration, an_rpc_response


def test_factory():
    switch = RealSwitchFactory().get_switch_by_descriptor(
                SwitchDescriptor(hostname='hostname', model='juniper_mx', username='username', password='password', port=22)
            )

    assert_that(switch, instance_of(Juniper))
    assert_that(switch.switch_descriptor.hostname, equal_to("hostname"))
    assert_that(switch.switch_descriptor.model, equal_to("juniper_mx"))
    assert_that(switch.switch_descriptor.username, equal_to("username"))
    assert_that(switch.switch_descriptor.password, equal_to("password"))
    assert_that(switch.switch_descriptor.port, equal_to(22))


class JuniperMXTest(unittest.TestCase):
    def setUp(self):
        self.switch = netconf(SwitchDescriptor(model='juniper_mx', hostname="toto"))

        self.netconf_mock = flexmock()
        self.switch.netconf = self.netconf_mock
        self.switch.in_transaction = True

    def tearDown(self):
        flexmock_teardown()

    def test_add_vlan(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <bridge-domains />
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>PATATE</name>
                <vlan-id>900</vlan-id>
              </domain>
            </bridge-domains>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <bridge-domains>
                  <domain>
                    <name>VLAN1000</name>
                    <vlan-id>1000</vlan-id>
                    <description>Shizzle</description>
                  </domain>
                </bridge-domains>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.add_vlan(1000, name="Shizzle")

    def test_add_vlan_already_in_use_raises(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <bridge-domains />
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>PATATE</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
        """))

        self.netconf_mock.should_receive("edit_config").never()

        with self.assertRaises(VlanAlreadyExist) as expect:
            self.switch.add_vlan(1000)

        assert_that(str(expect.exception), contains_string("Vlan 1000 already exist"))

    def test_add_existing_vlan_raises(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <bridge-domains />
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>VLAN1000</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
        """))

        self.netconf_mock.should_receive("edit_config").never()

        with self.assertRaises(VlanAlreadyExist) as expect:
            self.switch.add_vlan(1000)

        assert_that(str(expect.exception), contains_string("Vlan 1000 already exist"))

    def test_add_vlan_bad_vlan_id(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
                <filter>
                  <configuration>
                    <bridge-domains />
                  </configuration>
                </filter>
            """)).and_return(a_configuration(""))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
                <config>
                  <configuration>
                    <bridge-domains>
                      <domain>
                        <name>VLAN9000</name>
                        <vlan-id>9000</vlan-id>
                      </domain>
                    </bridge-domains>
                  </configuration>
                </config>
            """)).and_raise(RPCError(to_ele(textwrap.dedent("""
                <rpc-error xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" xmlns:junos="http://xml.juniper.net/junos/15.1R4/junos" xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0">>
                <error-severity>error</error-severity>
                <error-info>
                  <bad-element>domain</bad-element>
                </error-info>
                <error-message>Value 9000 is not within range (1..4094)</error-message>
                </rpc-error>
            """))))

        with self.assertRaises(BadVlanNumber) as expect:
            self.switch.add_vlan(9000)

        assert_that(str(expect.exception), equal_to("Vlan number is invalid"))

    def test_add_vlan_empty_vlan_name(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
                <filter>
                  <configuration>
                    <bridge-domains />
                  </configuration>
                </filter>
            """)).and_return(a_configuration(""))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
                <config>
                  <configuration>
                    <bridge-domains>
                      <domain>
                        <name>VLAN1000</name>
                        <vlan-id>1000</vlan-id>
                        <description></description>
                      </domain>
                    </bridge-domains>
                  </configuration>
                </config>
            """)).and_raise(RPCError(to_ele(textwrap.dedent("""
                 <rpc-error xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"
                 xmlns:junos="http://xml.juniper.net/junos/15.1R4/junos"
                 xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0">
                    <error-type>protocol</error-type>
                    <error-tag>operation-failed</error-tag>
                    <error-severity>error</error-severity>
                    <error-message>description: '': Must be a string of 255 characters or less</error-message>
                    <error-info>
                      <bad-element>domain</bad-element>
                    </error-info>
                  </rpc-error>
            """))))

        with self.assertRaises(BadVlanName) as expect:
            self.switch.add_vlan(1000, "")

        assert_that(str(expect.exception), equal_to("Vlan name is invalid"))

    def test_add_vlan_too_long_vlan_name(self):
        long_string = 'a' * 256
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
                <filter>
                  <configuration>
                    <bridge-domains />
                  </configuration>
                </filter>
            """)).and_return(a_configuration(""))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
                <config>
                  <configuration>
                    <bridge-domains>
                      <domain>
                        <name>VLAN1000</name>
                        <vlan-id>1000</vlan-id>
                        <description>{}</description>
                      </domain>
                    </bridge-domains>
                  </configuration>
                </config>
            """.format(long_string))).and_raise(RPCError(to_ele(textwrap.dedent("""
                 <rpc-error xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"
                 xmlns:junos="http://xml.juniper.net/junos/15.1R4/junos"
                 xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0">
                    <error-type>protocol</error-type>
                    <error-tag>operation-failed</error-tag>
                    <error-severity>error</error-severity>
                    <error-message>description: '{}': Must be a string of 255 characters or less</error-message>
                    <error-info>
                      <bad-element>domain</bad-element>
                    </error-info>
                  </rpc-error>
            """.format(long_string)))))

        with self.assertRaises(BadVlanName) as expect:
            self.switch.add_vlan(1000, long_string)

        assert_that(str(expect.exception), equal_to("Vlan name is invalid"))

    def test_add_vlan_raises_RPCError(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
                <filter>
                  <configuration>
                    <bridge-domains />
                  </configuration>
                </filter>
            """)).and_return(a_configuration(""))

        self.netconf_mock.should_receive("edit_config").once().and_raise(RPCError(to_ele(textwrap.dedent("""
                 <rpc-error xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"
                 xmlns:junos="http://xml.juniper.net/junos/15.1R4/junos"
                 xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0">
                    <error-type>protocol</error-type>
                    <error-tag>operation-failed</error-tag>
                    <error-severity>error</error-severity>
                    <error-message>There's another problem</error-message>
                    <error-info>
                      <bad-element>domain</bad-element>
                    </error-info>
                  </rpc-error>
            """))))

        with self.assertRaises(RPCError):
            self.switch.add_vlan(1000, 'a' * 256)

    def test_remove_vlan_ignores_removing_interface_not_created(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <bridge-domains />
                <interfaces />
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>STANDARD</name>
                <vlan-id>10</vlan-id>
              </domain>
            </bridge-domains>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <bridge-domains>
                  <domain operation="delete">
                    <name>STANDARD</name>
                  </domain>
                </bridge-domains>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.remove_vlan(10)

    def test_remove_vlan_invalid_vlan_raises(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <bridge-domains />
                <interfaces />
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>ANOTHER</name>
                <vlan-id>10</vlan-id>
              </domain>
            </bridge-domains>
        """))

        with self.assertRaises(UnknownVlan) as expect:
            self.switch.remove_vlan(20)

        assert_that(str(expect.exception), equal_to("Vlan 20 not found"))

    def test_remove_vlan_also_removes_associated_interface(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <bridge-domains/>
                <interfaces/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>MEH</name>
                <vlan-id>5</vlan-id>
              </domain>
              <domain>
                <name>STANDARD</name>
                <vlan-id>10</vlan-id>
                <routing-interface>irb.25</routing-interface>
              </domain>
              <domain>
                <name>MEH2</name>
                <vlan-id>15</vlan-id>
              </domain>
            </bridge-domains>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <bridge-domains>
                  <domain operation="delete">
                    <name>STANDARD</name>
                  </domain>
                </bridge-domains>
                <interfaces>
                  <interface>
                    <name>irb</name>
                    <unit operation="delete">
                      <name>25</name>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.remove_vlan(10)

    def test_remove_vlan_in_use_deletes_all_usages(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <bridge-domains />
                <interfaces />
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>STANDARD</name>
                <vlan-id>10</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/1</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>trunk</interface-mode>
                        <vlan-id-list>9</vlan-id-list>
                        <vlan-id-list>10</vlan-id-list>
                        <vlan-id-list>11</vlan-id-list>
                    </bridge>
                  </family>
                </unit>
              </interface>
              <interface>
                <name>xe-0/0/2</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>trunk</interface-mode>
                        <vlan-id-list>9-15</vlan-id-list>
                    </bridge>
                  </family>
                </unit>
              </interface>
              <interface>
                <name>xe-0/0/3</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                        <interface-mode>access</interface-mode>
                        <vlan-id-list>12</vlan-id-list>
                    </bridge>
                  </family>
                </unit>
              </interface>
              <interface>
                <name>xe-0/0/4</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>access</interface-mode>
                        <vlan-id>STANDARD</vlan-id>
                    </bridge>
                  </family>
                </unit>
              </interface>
              <interface>
                <name>xe-0/0/5</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>access</interface-mode>
                        <vlan-id>ANOTHER_NAME</vlan-id>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <bridge-domains>
                  <domain operation="delete">
                    <name>STANDARD</name>
                  </domain>
                </bridge-domains>
                <interfaces>
                  <interface>
                    <name>xe-0/0/1</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                          <vlan-id-list operation="delete">10</vlan-id-list>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                  <interface>
                    <name>xe-0/0/2</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                            <vlan-id-list operation="delete">9-15</vlan-id-list>
                            <vlan-id-list>9</vlan-id-list>
                            <vlan-id-list>11-15</vlan-id-list>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                  <interface>
                    <name>xe-0/0/4</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                            <vlan-id operation="delete">STANDARD</vlan-id>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>""")).and_return(an_ok_response())

        self.switch.remove_vlan(10)

    def test_remove_vlan_delete_usage_and_interface_at_same_time(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <bridge-domains />
                <interfaces />
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>STANDARD</name>
                <vlan-id>10</vlan-id>
                <routing-interface>irb.10</routing-interface>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>name</name>
                <unit>
                  <name>10</name>
                  <family>
                    <inet>
                      <address>
                        <name>1.1.1.1/24</name>
                      </address>
                    </inet>
                  </family>
                </unit>
              </interface>
              <interface>
                <name>xe-0/0/1</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <port-mode>trunk</port-mode>
                        <vlan-id-list>10</vlan-id-list>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <bridge-domains>
                  <domain operation="delete">
                    <name>STANDARD</name>
                  </domain>
                </bridge-domains>
                <interfaces>
                  <interface>
                    <name>irb</name>
                    <unit operation="delete">
                      <name>10</name>
                    </unit>
                  </interface>
                  <interface>
                    <name>xe-0/0/1</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                            <vlan-id-list operation="delete">10</vlan-id-list>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.remove_vlan(10)

    def test_get_vlans(self):
        self.switch.in_transaction = False

        self.netconf_mock.should_receive("get_config").with_args(source="running", filter=is_xml("""
                <filter>
                  <configuration>
                    <bridge-domains />
                    <interfaces />
                  </configuration>
                </filter>
            """)).and_return(a_configuration("""
                <bridge-domains>
                  <domain>
                    <name>STANDARD</name>
                    <vlan-id>10</vlan-id>
                    <description>my-description</description>
                  </domain>
                  <domain>
                    <name>NO-VLAN-ID</name>
                    <description>shizzle</description>
                  </domain>
                  <domain>
                    <name>WITH-IF</name>
                    <vlan-id>20</vlan-id>
                    <routing-interface>irb.20</routing-interface>
                  </domain>
                  <domain>
                    <name>WITH-IF-MULTI-IP</name>
                    <vlan-id>40</vlan-id>
                    <routing-interface>irb.70</routing-interface>
                  </domain>
                </bridge-domains>
                <interfaces>
                  <interface>
                    <name>xe-0/0/1</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                  <interface>
                    <name>irb</name>
                    <unit>
                      <name>20</name>
                      <family>
                        <inet>
                          <address>
                            <name>1.1.1.1/24</name>
                          </address>
                          <filter>
                            <input>
                              <filter-name>AC-IN</filter-name>
                            </input>
                            <output>
                              <filter-name>AC-OUT</filter-name>
                            </output>
                          </filter>
                        </inet>
                      </family>
                    </unit>
                    <unit>
                      <name>40</name>
                    </unit>
                    <unit>
                      <name>70</name>
                      <family>
                        <inet>
                          <address>
                            <name>2.1.1.1/24</name>
                          </address>
                          <address>
                            <name>4.1.1.1/24</name>
                          </address>
                          <address>
                            <name>3.1.1.1/24</name>
                          </address>
                        </inet>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
            """))

        vlan10, vlan20, vlan40 = self.switch.get_vlans()

        assert_that(vlan10.number, equal_to(10))
        assert_that(vlan10.name, equal_to("my-description"))
        assert_that(vlan10.access_groups[IN], equal_to(None))
        assert_that(vlan10.access_groups[OUT], equal_to(None))
        assert_that(vlan10.ips, has_length(0))

        assert_that(vlan20.number, equal_to(20))
        assert_that(vlan20.name, equal_to(None))
        assert_that(vlan20.access_groups[IN], equal_to("AC-IN"))
        assert_that(vlan20.access_groups[OUT], equal_to("AC-OUT"))
        assert_that(vlan20.ips, has_length(1))
        vlan20ip1 = vlan20.ips[0]
        assert_that(str(vlan20ip1.ip), equal_to("1.1.1.1"))
        assert_that(vlan20ip1.prefixlen, equal_to(24))

        assert_that(vlan40.number, equal_to(40))
        assert_that(vlan40.name, equal_to(None))
        assert_that(vlan40.access_groups[IN], equal_to(None))
        assert_that(vlan40.access_groups[OUT], equal_to(None))
        vlan40ip1, vlan40ip2, vlan40ip3 = vlan40.ips
        assert_that(str(vlan40ip1.ip), equal_to("2.1.1.1"))
        assert_that(vlan40ip1.prefixlen, equal_to(24))
        assert_that(str(vlan40ip2.ip), equal_to("3.1.1.1"))
        assert_that(vlan40ip2.prefixlen, equal_to(24))
        assert_that(str(vlan40ip3.ip), equal_to("4.1.1.1"))
        assert_that(vlan40ip3.prefixlen, equal_to(24))

    def test_get_vlan_with_interface_multi_ip(self):
        self.switch.in_transaction = False
        self.netconf_mock.should_receive("get_config").with_args(source="running", filter=is_xml("""
                <filter>
                  <configuration>
                    <bridge-domains />
                    <interfaces />
                  </configuration>
                </filter>
            """)).and_return(a_configuration("""
                <bridge-domains>
                  <domain>
                    <name>This-another-clam</name>
                    <vlan-id>39</vlan-id>
                    <routing-interface>irb.20</routing-interface>
                  </domain>
                  <domain>
                    <name>WITH-IF-MULTI-IP</name>
                    <vlan-id>40</vlan-id>
                    <routing-interface>irb.70</routing-interface>
                  </domain>
                  <domain>
                    <name>This-yet-another-clam</name>
                    <vlan-id>41</vlan-id>
                    <routing-interface>irb.40</routing-interface>
                  </domain>
                </bridge-domains>
                <interfaces>
                  <interface>
                    <name>xe-0/0/1</name>
                  </interface>
                  <interface>
                    <name>irb</name>
                    <unit>
                      <name>20</name>
                      <family>
                        <inet>
                          <address>
                            <name>1.1.1.1/24</name>
                          </address>
                          <filter>
                            <input>
                              <filter-name>AC-IN</filter-name>
                            </input>
                            <output>
                              <filter-name>AC-OUT</filter-name>
                            </output>
                          </filter>
                        </inet>
                      </family>
                    </unit>
                    <unit>
                      <name>40</name>
                    </unit>
                    <unit>
                      <name>70</name>
                      <family>
                        <inet>
                          <address>
                            <name>2.1.1.1/24</name>
                          </address>
                          <address>
                            <name>4.1.1.1/24</name>
                          </address>
                          <address>
                            <name>3.1.1.1/24</name>
                          </address>
                        </inet>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
            """))

        vlan = self.switch.get_vlan(40)

        assert_that(vlan.number, equal_to(40))
        assert_that(vlan.name, equal_to(None))
        assert_that(vlan.access_groups[IN], equal_to(None))
        assert_that(vlan.access_groups[OUT], equal_to(None))
        assert_that(vlan.icmp_redirects, equal_to(True))
        vlanip1, vlanip2, vlanip3 = vlan.ips
        assert_that(str(vlanip1.ip), equal_to("2.1.1.1"))
        assert_that(vlanip1.prefixlen, equal_to(24))
        assert_that(str(vlanip2.ip), equal_to("3.1.1.1"))
        assert_that(vlanip2.prefixlen, equal_to(24))
        assert_that(str(vlanip3.ip), equal_to("4.1.1.1"))
        assert_that(vlanip3.prefixlen, equal_to(24))

    def test_get_vlan_with_no_interface(self):
        self.switch.in_transaction = False
        self.netconf_mock.should_receive("get_config").with_args(source="running", filter=is_xml("""
                <filter>
                  <configuration>
                    <bridge-domains />
                    <interfaces />
                  </configuration>
                </filter>
            """)).and_return(a_configuration("""
                <bridge-domains>
                  <domain>
                    <name>STANDARD</name>
                    <vlan-id>10</vlan-id>
                    <description>my-description</description>
                  </domain>
                </bridge-domains>
            """))

        vlan = self.switch.get_vlan(10)

        assert_that(vlan.number, equal_to(10))
        assert_that(vlan.name, equal_to("my-description"))
        assert_that(vlan.access_groups[IN], equal_to(None))
        assert_that(vlan.access_groups[OUT], equal_to(None))
        assert_that(vlan.icmp_redirects, equal_to(True))
        assert_that(vlan.ips, has_length(0))

    def test_get_vlan_with_unknown_vlan(self):
        self.switch.in_transaction = False
        self.netconf_mock.should_receive("get_config").with_args(source="running", filter=is_xml("""
                <filter>
                  <configuration>
                    <bridge-domains />
                    <interfaces />
                  </configuration>
                </filter>
            """)).and_return(a_configuration("""
                <bridge-domains>
                  <domain>
                    <name>This-another-clam</name>
                    <vlan-id>39</vlan-id>
                  </domain>
                </bridge-domains>
            """))

        with self.assertRaises(UnknownVlan) as expect:
            self.switch.get_vlan(10)

        assert_that(str(expect.exception), equal_to("Vlan 10 not found"))

    def test_get_vlan_with_interface(self):
        self.switch.in_transaction = False
        self.netconf_mock.should_receive("get_config").with_args(source="running", filter=is_xml("""
                <filter>
                  <configuration>
                    <bridge-domains />
                    <interfaces />
                  </configuration>
                </filter>
            """)).and_return(a_configuration("""
                <bridge-domains>
                  <domain>
                    <name>WITH-IF</name>
                    <vlan-id>20</vlan-id>
                    <routing-interface>irb.20</routing-interface>
                  </domain>
                </bridge-domains>
                <interfaces>
              <interface>
                <name>xe-0/0/1</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                    </bridge>
                  </family>
                </unit>
              </interface>
                  <interface>
                    <name>xe-0/0/1</name>
                  </interface>
                  <interface>
                    <name>irb</name>
                    <unit>
                      <name>20</name>
                      <family>
                        <inet>
                          <no-redirects />
                          <address>
                            <name>1.1.1.1/24</name>
                          </address>
                          <filter>
                            <input>
                              <filter-name>AC-IN</filter-name>
                            </input>
                            <output>
                              <filter-name>AC-OUT</filter-name>
                            </output>
                          </filter>
                        </inet>
                      </family>
                    </unit>
                    <unit>
                      <name>40</name>
                    </unit>
                    <unit>
                      <name>70</name>
                      <family>
                        <inet>
                          <address>
                            <name>2.1.1.1/24</name>
                          </address>
                          <address>
                            <name>4.1.1.1/24</name>
                          </address>
                          <address>
                            <name>3.1.1.1/24</name>
                          </address>
                        </inet>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
            """))

        vlan = self.switch.get_vlan(20)

        assert_that(vlan.number, equal_to(20))
        assert_that(vlan.name, equal_to(None))
        assert_that(vlan.access_groups[IN], equal_to("AC-IN"))
        assert_that(vlan.access_groups[OUT], equal_to("AC-OUT"))
        assert_that(vlan.icmp_redirects, equal_to(False))
        assert_that(vlan.ips, has_length(1))
        vlan20ip1 = vlan.ips[0]
        assert_that(str(vlan20ip1.ip), equal_to("1.1.1.1"))
        assert_that(vlan20ip1.prefixlen, equal_to(24))

    def test_get_vlan_with_vrrp(self):
        self.switch.in_transaction = False
        self.netconf_mock.should_receive("get_config").with_args(source="running", filter=is_xml("""
                <filter>
                  <configuration>
                    <bridge-domains />
                    <interfaces />
                  </configuration>
                </filter>
            """)).and_return(a_configuration("""
                <bridge-domains>
                  <domain>
                    <name>WITH-IF</name>
                    <vlan-id>20</vlan-id>
                    <routing-interface>irb.20</routing-interface>
                  </domain>
                </bridge-domains>
                <interfaces>
                <interface>
                  <name>irb</name>
                  <unit>
                    <name>20</name>
                    <family>
                      <inet>
                        <address>
                          <name>1.1.1.2/24</name>
                          <vrrp-group>
                            <name>1</name>
                            <virtual-address>1.1.1.1</virtual-address>
                            <priority>90</priority>
                            <preempt>
                              <hold-time>60</hold-time>
                            </preempt>
                            <accept-data/>
                            <authentication-type>simple</authentication-type>
                            <authentication-key>$9$1/aElvwsgoaGz3reKvLX.Pf5n/</authentication-key>
                            <track>
                              <route>
                                <route_address>0.0.0.0/0</route_address>
                                <routing-instance>default</routing-instance>
                                <priority-cost>50</priority-cost>
                              </route>
                            </track>
                        </vrrp-group>
                        </address>
                      </inet>
                    </family>
                  </unit>
                </interface>
                </interfaces>
            """))

        vlan = self.switch.get_vlan(20)

        vrrp = vlan.vrrp_groups[0]
        assert_that(vrrp.id, is_(1))
        assert_that(vrrp.ips, has_length(1))
        assert_that(vrrp.ips[0], is_(IPAddress('1.1.1.1')))
        assert_that(vrrp.priority, is_(90))
        assert_that(vrrp.hello_interval, is_(None))
        assert_that(vrrp.dead_interval, is_(None))
        assert_that(vrrp.track_id, is_("0.0.0.0/0"))
        assert_that(vrrp.track_decrement, is_(50))

    def test_get_vlan_with_vrrp_without_optional_fields_and_multiple_vips(self):
        self.switch.in_transaction = False
        self.netconf_mock.should_receive("get_config").with_args(source="running", filter=is_xml("""
                <filter>
                  <configuration>
                    <bridge-domains />
                    <interfaces />
                  </configuration>
                </filter>
            """)).and_return(a_configuration("""
                <bridge-domains>
                  <domain>
                    <name>WITH-IF</name>
                    <vlan-id>20</vlan-id>
                    <routing-interface>irb.20</routing-interface>
                  </domain>
                </bridge-domains>
                <interfaces>
                <interface>
                  <name>irb</name>
                  <unit>
                    <name>20</name>
                    <family>
                      <inet>
                        <address>
                          <name>1.1.1.2/24</name>
                          <vrrp-group>
                            <name>1</name>
                            <virtual-address>1.1.1.1</virtual-address>
                            <virtual-address>1.1.1.3</virtual-address>
                          </vrrp-group>
                        </address>
                      </inet>
                    </family>
                  </unit>
                </interface>
                </interfaces>
            """))

        vlan = self.switch.get_vlan(20)

        vrrp = vlan.vrrp_groups[0]
        assert_that(vrrp.id, is_(1))
        assert_that(vrrp.ips, has_length(2))
        assert_that(vrrp.ips[0], is_(IPAddress('1.1.1.1')))
        assert_that(vrrp.ips[1], is_(IPAddress('1.1.1.3')))
        assert_that(vrrp.priority, is_(None))
        assert_that(vrrp.hello_interval, is_(None))
        assert_that(vrrp.dead_interval, is_(None))
        assert_that(vrrp.track_id, is_(None))
        assert_that(vrrp.track_decrement, is_(None))

    def test_add_vrrp_success(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                      <unit>
                        <name>1234</name>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>irb</name>
                  <unit>
                    <name>1234</name>
                    <family>
                      <inet>
                        <address>
                            <name>3.3.3.2/27</name>
                        </address>
                      </inet>
                    </family>
                  </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                      <unit>
                        <name>1234</name>
                        <family>
                          <inet>
                            <address>
                              <name>3.3.3.2/27</name>
                              <vrrp-group>
                                <name>1</name>
                                <priority>110</priority>
                                <preempt>
                                  <hold-time>60</hold-time>
                                </preempt>
                                <accept-data/>
                                <authentication-type>simple</authentication-type>
                                <authentication-key>VLAN1234</authentication-key>
                                <track>
                                  <route>
                                    <route_address>0.0.0.0/0</route_address>
                                    <routing-instance>default</routing-instance>
                                    <priority-cost>50</priority-cost>
                                  </route>
                                </track>
                                <virtual-address>3.3.3.1</virtual-address>
                              </vrrp-group>
                            </address>
                          </inet>
                        </family>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>""")).and_return(an_ok_response())

        self.switch.add_vrrp_group(1234, 1, ips=[IPAddress("3.3.3.1")], priority=110, track_id="0.0.0.0/0",
                                   track_decrement=50)

    def test_add_vrrp_multiple_ips(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                      <unit>
                        <name>1234</name>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>irb</name>
                  <unit>
                    <name>1234</name>
                    <family>
                      <inet>
                        <address>
                            <name>3.3.3.2/27</name>
                        </address>
                      </inet>
                    </family>
                  </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                      <unit>
                        <name>1234</name>
                        <family>
                          <inet>
                            <address>
                              <name>3.3.3.2/27</name>
                              <vrrp-group>
                                <name>1</name>
                                <priority>110</priority>
                                <preempt>
                                  <hold-time>60</hold-time>
                                </preempt>
                                <accept-data/>
                                <authentication-type>simple</authentication-type>
                                <authentication-key>VLAN1234</authentication-key>
                                <track>
                                  <route>
                                    <route_address>0.0.0.0/0</route_address>
                                    <routing-instance>default</routing-instance>
                                    <priority-cost>50</priority-cost>
                                  </route>
                                </track>
                                <virtual-address>3.3.3.1</virtual-address>
                                <virtual-address>3.3.3.3</virtual-address>
                              </vrrp-group>
                            </address>
                          </inet>
                        </family>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>""")).and_return(an_ok_response())

        self.switch.add_vrrp_group(1234, 1, ips=[IPAddress("3.3.3.1"), IPAddress("3.3.3.3")], priority=110,
                                   track_id="0.0.0.0/0", track_decrement=50)

    def test_add_vrrp_fails_when_vlan_not_found(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                      <unit>
                        <name>1234</name>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration())

        with self.assertRaises(UnknownVlan):
            self.switch.add_vrrp_group(1234, 1, ips=[IPAddress("3.3.3.1")], priority=110, track_id="0.0.0.0/0",
                                       track_decrement=50)

    def test_add_vrrp_adds_it_to_the_good_address(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                      <unit>
                        <name>1234</name>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>irb</name>
                  <unit>
                    <name>1234</name>
                    <family>
                      <inet>
                        <address>
                            <name>3.3.3.2/27</name>
                        </address>
                        <address>
                            <name>4.4.4.2/27</name>
                        </address>
                        <address>
                            <name>5.5.5.2/27</name>
                        </address>
                      </inet>
                    </family>
                  </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                      <unit>
                        <name>1234</name>
                        <family>
                          <inet>
                            <address>
                              <name>4.4.4.2/27</name>
                              <vrrp-group>
                                <name>1</name>
                                <priority>110</priority>
                                <preempt>
                                  <hold-time>60</hold-time>
                                </preempt>
                                <accept-data/>
                                <authentication-type>simple</authentication-type>
                                <authentication-key>VLAN1234</authentication-key>
                                <track>
                                  <route>
                                    <route_address>0.0.0.0/0</route_address>
                                    <routing-instance>default</routing-instance>
                                    <priority-cost>50</priority-cost>
                                  </route>
                                </track>
                                <virtual-address>4.4.4.1</virtual-address>
                              </vrrp-group>
                            </address>
                          </inet>
                        </family>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>""")).and_return(an_ok_response())

        self.switch.add_vrrp_group(1234, 1, ips=[IPAddress("4.4.4.1")], priority=110, track_id="0.0.0.0/0",
                                   track_decrement=50)

    def test_add_vrrp_adds_it_if_all_ips_are_within_a_single_address(self):
        self.netconf_mock.should_receive("get_config").and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>irb</name>
                  <unit>
                    <name>1234</name>
                    <family>
                      <inet>
                        <address>
                            <name>1.1.1.1/27</name>
                        </address>
                      </inet>
                    </family>
                  </unit>
              </interface>
            </interfaces>
        """))
        self.netconf_mock.should_receive("edit_config").once()
        self.switch.add_vrrp_group(1234, 1, ips=[IPAddress("1.1.1.2"), IPAddress("1.1.1.3")],
                                   priority=110, track_id="0.0.0.0/0",
                                   track_decrement=50)

    def test_add_vrrp_fails_when_the_ips_doesnt_belong_to_an_existing_address(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                      <unit>
                        <name>1234</name>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>irb</name>
                  <unit>
                    <name>1234</name>
                    <family>
                      <inet>
                        <address>
                            <name>3.3.3.2/27</name>
                        </address>
                      </inet>
                    </family>
                  </unit>
              </interface>
            </interfaces>
        """))

        with self.assertRaises(UnknownIP):
            self.switch.add_vrrp_group(1234, 1, ips=[IPAddress("4.4.4.1")], priority=110, track_id="0.0.0.0/0",
                                       track_decrement=50)

    def test_add_vrrp_fail_if_all_ips_are_not_in_the_same_address(self):
        self.netconf_mock.should_receive("get_config").and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>irb</name>
                  <unit>
                    <name>1234</name>
                    <family>
                      <inet>
                        <address>
                            <name>1.1.1.1/27</name>
                        </address>
                        <address>
                            <name>2.2.2.2/27</name>
                        </address>
                      </inet>
                    </family>
                  </unit>
              </interface>
            </interfaces>
        """))
        with self.assertRaises(UnknownIP):
            self.switch.add_vrrp_group(1234, 1, ips=[IPAddress("1.1.1.2"), IPAddress("2.2.2.3")],
                                       priority=110, track_id="0.0.0.0/0",
                                       track_decrement=50)

    def test_add_vrrp_fails_when_any_of_the_ips_doesnt_belong_to_an_existing_address(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                      <unit>
                        <name>1234</name>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>irb</name>
                  <unit>
                    <name>1234</name>
                    <family>
                      <inet>
                        <address>
                            <name>3.3.3.2/27</name>
                        </address>
                      </inet>
                    </family>
                  </unit>
              </interface>
            </interfaces>
        """))

        with self.assertRaises(UnknownIP):
            self.switch.add_vrrp_group(1234, 1, ips=[IPAddress("3.3.3.1"), IPAddress("4.4.4.1")], priority=110, track_id="0.0.0.0/0",
                                       track_decrement=50)

    def test_remove_vrrp_success(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                      <unit>
                        <name>1234</name>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>irb</name>
                  <unit>
                    <name>1234</name>
                    <family>
                      <inet>
                        <address>
                          <name>192.0.1.1/27</name>
                          <vrrp-group>
                            <name>1</name>
                            <virtual-address>192.0.1.2</virtual-address>
                          </vrrp-group>
                        </address>
                      </inet>
                    </family>
                  </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                    <unit>
                      <name>1234</name>
                      <family>
                        <inet>
                          <address>
                            <name>192.0.1.1/27</name>
                            <vrrp-group operation="delete">
                              <name>1</name>
                            </vrrp-group>
                          </address>
                        </inet>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>""")).and_return(an_ok_response())

        self.switch.remove_vrrp_group(1234, 1)

    def test_remove_vrrp_with_invalid_group_id(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                      <unit>
                        <name>1234</name>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>irb</name>
                  <unit>
                    <name>1234</name>
                    <family>
                      <inet>
                        <address>
                          <name>192.0.1.1/27</name>
                          <vrrp-group>
                            <name>99</name>
                            <virtual-address>192.0.1.2</virtual-address>
                          </vrrp-group>
                        </address>
                      </inet>
                    </family>
                  </unit>
              </interface>
            </interfaces>"""))

        self.netconf_mock.should_receive("edit_config").never()

        with self.assertRaises(VrrpDoesNotExistForVlan) as expect:
            self.switch.remove_vrrp_group(1234, 1)

        assert_that(str(expect.exception), equal_to("Vrrp group 1 does not exist for vlan 1234"))

    def test_remove_vrrp_from_unknown_vlan(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                      <unit>
                        <name>1234</name>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration())

        self.netconf_mock.should_receive("edit_config").never()

        with self.assertRaises(UnknownVlan) as expect:
            self.switch.remove_vrrp_group(1234, 2)

        assert_that(str(expect.exception), equal_to("Vlan 1234 not found"))

    def test_add_ip_to_vlan(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <bridge-domains>
                  <domain>
                    <vlan-id>1234</vlan-id>
                  </domain>
                </bridge-domains>
                <interfaces>
                  <interface>
                    <name>irb</name>
                    <unit>
                      <name>1234</name>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>VLAN1234</name>
                <vlan-id>1234</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces/>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <bridge-domains>
                  <domain>
                    <name>VLAN1234</name>
                    <vlan-id>1234</vlan-id>
                    <routing-interface>irb.1234</routing-interface>
                  </domain>
                </bridge-domains>
                <interfaces>
                  <interface>
                    <name>irb</name>
                    <unit>
                      <name>1234</name>
                      <family>
                        <inet>
                          <address>
                            <name>3.3.3.2/27</name>
                          </address>
                        </inet>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.add_ip_to_vlan(vlan_number=1234, ip_network=IPNetwork("3.3.3.2/27"))

    def test_add_ip_to_vlan_unknown_vlan_raises(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <bridge-domains>
                  <domain>
                    <vlan-id>1234</vlan-id>
                  </domain>
                </bridge-domains>
                <interfaces>
                  <interface>
                    <name>irb</name>
                    <unit>
                      <name>1234</name>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration())

        with self.assertRaises(UnknownVlan):
            self.switch.add_ip_to_vlan(vlan_number=1234, ip_network=IPNetwork("3.3.3.2/27"))

    def test_add_ip_to_vlan_ip_already_exists_in_vlan_raises(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <bridge-domains>
                  <domain>
                    <vlan-id>1234</vlan-id>
                  </domain>
                </bridge-domains>
                <interfaces>
                  <interface>
                    <name>irb</name>
                    <unit>
                      <name>1234</name>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>VLAN1234</name>
                <vlan-id>1234</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>irb</name>
                  <unit>
                    <name>1234</name>
                    <family>
                      <inet>
                        <address>
                            <name>3.3.3.2/27</name>
                        </address>
                      </inet>
                    </family>
                  </unit>
              </interface>
            </interfaces>
        """))

        with self.assertRaises(IPAlreadySet):
            self.switch.add_ip_to_vlan(vlan_number=1234, ip_network=IPNetwork("3.3.3.2/27"))

    def test_set_icmp_redirect_state_false_not_set_adds_statement(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <bridge-domains>
                  <domain>
                    <vlan-id>1234</vlan-id>
                  </domain>
                </bridge-domains>
                <interfaces>
                  <interface>
                    <name>irb</name>
                    <unit>
                      <name>1234</name>
                      <family>
                        <inet>
                          <no-redirects />
                        </inet>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>VLAN1234</name>
                <vlan-id>1234</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces/>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <bridge-domains>
                  <domain>
                    <name>VLAN1234</name>
                    <vlan-id>1234</vlan-id>
                    <routing-interface>irb.1234</routing-interface>
                  </domain>
                </bridge-domains>
                <interfaces>
                  <interface>
                    <name>irb</name>
                    <unit>
                      <name>1234</name>
                      <family>
                        <inet>
                          <no-redirects />
                        </inet>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.set_vlan_icmp_redirects_state(vlan_number=1234, state=False)

    def test_set_icmp_redirect_state_false_already_set_dont_do_anything(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <bridge-domains>
                  <domain>
                    <vlan-id>1234</vlan-id>
                  </domain>
                </bridge-domains>
                <interfaces>
                  <interface>
                    <name>irb</name>
                    <unit>
                      <name>1234</name>
                      <family>
                        <inet>
                          <no-redirects />
                        </inet>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>VLAN1234</name>
                <vlan-id>1234</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>irb</name>
                <unit>
                  <name>4094</name>
                  <family>
                    <inet>
                      <no-redirects/>
                    </inet>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").never()

        self.switch.set_vlan_icmp_redirects_state(vlan_number=1234, state=False)

    def test_set_icmp_redirect_state_true_not_set_dont_do_anything(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <bridge-domains>
                  <domain>
                    <vlan-id>1234</vlan-id>
                  </domain>
                </bridge-domains>
                <interfaces>
                  <interface>
                    <name>irb</name>
                    <unit>
                      <name>1234</name>
                      <family>
                        <inet>
                          <no-redirects />
                        </inet>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>VLAN1234</name>
                <vlan-id>1234</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces/>
        """))

        self.netconf_mock.should_receive("edit_config").never()

        self.switch.set_vlan_icmp_redirects_state(vlan_number=1234, state=True)

    def test_set_icmp_redirect_state_true_already_set_remove_statement(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <bridge-domains>
                  <domain>
                    <vlan-id>1234</vlan-id>
                  </domain>
                </bridge-domains>
                <interfaces>
                  <interface>
                    <name>irb</name>
                    <unit>
                      <name>1234</name>
                      <family>
                        <inet>
                          <no-redirects />
                        </inet>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>VLAN1234</name>
                <vlan-id>1234</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>irb</name>
                <unit>
                  <name>4094</name>
                  <family>
                    <inet>
                      <no-redirects/>
                    </inet>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                    <unit>
                      <name>1234</name>
                      <family>
                        <inet>
                          <no-redirects operation="delete" />
                        </inet>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """))

        self.switch.set_vlan_icmp_redirects_state(vlan_number=1234, state=True)

    def test_set_icmp_redirect_unknow_vlan_raises(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <bridge-domains>
                  <domain>
                    <vlan-id>1234</vlan-id>
                  </domain>
                </bridge-domains>
                <interfaces>
                  <interface>
                    <name>irb</name>
                    <unit>
                      <name>1234</name>
                      <family>
                        <inet>
                          <no-redirects />
                        </inet>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration())

        with self.assertRaises(UnknownVlan):
            self.switch.set_vlan_icmp_redirects_state(vlan_number=1234, state=False)

    def test_port_mode_access(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
            <bridge-domains/>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/6</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                          <interface-mode>access</interface-mode>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.set_access_mode("xe-0/0/6")

    def test_port_mode_access_with_no_mode_and_1_vlan_does_not_remove_it(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <vlan-id>2998</vlan-id>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
            <bridge-domains/>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/6</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                          <interface-mode>access</interface-mode>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.set_access_mode("xe-0/0/6")

    def test_set_access_vlan_on_interface_with_access_mode_and_no_vlan_succeeds_easily(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>PATATE</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>access</interface-mode>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/6</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                          <vlan-id>1000</vlan-id>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.set_access_vlan("xe-0/0/6", 1000)

    def test_set_access_vlan_on_interface_that_already_has_it_does_nothing(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>PATATE</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>access</interface-mode>
                      <vlan-id>1000</vlan-id>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").never()

        self.switch.set_access_vlan("xe-0/0/6", 1000)

    def test_set_access_vlan_on_interface_that_has_no_port_mode_sets_it(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>PATATE</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/6</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                          <interface-mode>access</interface-mode>
                            <vlan-id>1000</vlan-id>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.set_access_vlan("xe-0/0/6", 1000)

    def test_set_access_vlan_on_interface_replaces_vlan_id(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>PATATE</name>
                <vlan-id>1000</vlan-id>
              </domain>
              <domain>
                <name>PATATE2</name>
                <vlan-id>2000</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>access</interface-mode>
                      <vlan-id>2000</vlan-id>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/6</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                          <vlan-id>1000</vlan-id>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.set_access_vlan("xe-0/0/6", 1000)

    def test_set_access_vlan_on_interface_in_trunk_mode_should_raise(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>PATATE</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>trunk</interface-mode>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").never()

        with self.assertRaises(InterfaceInWrongPortMode) as expect:
            self.switch.set_access_vlan("xe-0/0/6", 1000)

        assert_that(str(expect.exception), contains_string("Operation cannot be performed on a trunk mode interface"))

    def test_set_access_vlan_on_unknown_vlan_raises(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>PATATE</name>
                <vlan-id>3333</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>access</interface-mode>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").never()

        with self.assertRaises(UnknownVlan) as expect:
            self.switch.set_access_vlan("xe-0/0/6", 1000)

        assert_that(str(expect.exception), contains_string("Vlan 1000 not found"))

    def test_set_access_vlan_on_default_interface_works(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>PATATE</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/6</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                          <interface-mode>access</interface-mode>
                            <vlan-id>1000</vlan-id>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.set_access_vlan("xe-0/0/6", 1000)

    def test_set_access_mode_on_interface_replaces_trunk_info(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>PATATE</name>
                <vlan-id>1000</vlan-id>
              </domain>
              <domain>
                <name>PATATE2</name>
                <vlan-id>2000</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>trunk</interface-mode>
                      <vlan-id-list>2000</vlan-id-list>
                      <vlan-id-list>2000-2000</vlan-id-list>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/6</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                          <interface-mode>access</interface-mode>
                          <vlan-id-list operation="delete"/>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.set_access_mode("xe-0/0/6")

    def test_port_mode_trunk_with_no_port_mode_or_vlan_set_just_sets_the_port_mode(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/6</name>
                  </interface>
                </interfaces>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
            <bridge-domains/>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/6</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                          <interface-mode>trunk</interface-mode>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.set_trunk_mode("xe-0/0/6")

    def test_set_port_mode_trunk_from_access_removes_vlan_info(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/6</name>
                  </interface>
                </interfaces>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>PATATE</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>access</interface-mode>
                      <vlan-id>1000</vlan-id>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/6</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                          <interface-mode>trunk</interface-mode>
                          <vlan-id operation="delete"/>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.set_trunk_mode("xe-0/0/6")

    def test_port_mode_trunk_already_in_trunk_mode_does_nothing(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/6</name>
                  </interface>
                </interfaces>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>trunk</interface-mode>
                      <vlan-id-list>1000</vlan-id-list>
                      <vlan-id-list>1001</vlan-id-list>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
            <vlans/>
        """))

        self.netconf_mock.should_receive("edit_config").never()

        self.switch.set_trunk_mode("xe-0/0/6")

    def test_add_trunk_vlan_on_interface_adds_to_the_list(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>PATATE</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>trunk</interface-mode>
                      <vlan-id-list>2000</vlan-id-list>
                      <vlan-id-list>2100-2200</vlan-id-list>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/6</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                          <vlan-id-list>1000</vlan-id-list>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.add_trunk_vlan("xe-0/0/6", 1000)

    def test_add_trunk_vlan_on_interface_that_has_no_port_mode_and_no_vlan_sets_it(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>PATATE</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/6</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                          <interface-mode>trunk</interface-mode>
                          <vlan-id-list>1000</vlan-id-list>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.add_trunk_vlan("xe-0/0/6", 1000)

    def test_add_trunk_vlan_on_interface_in_access_mode_raises(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>PATATE</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <vlan-id-list>500</vlan-id-list>
                      <interface-mode>access</interface-mode>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").never()

        with self.assertRaises(InterfaceInWrongPortMode) as expect:
            self.switch.add_trunk_vlan("xe-0/0/6", 1000)

        assert_that(str(expect.exception), contains_string("Operation cannot be performed on a access mode interface"))

    def test_add_trunk_vlan_on_unknown_vlan_raises(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>trunk</interface-mode>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").never()

        with self.assertRaises(UnknownVlan) as expect:
            self.switch.add_trunk_vlan("xe-0/0/6", 1000)

        assert_that(str(expect.exception), contains_string("Vlan 1000 not found"))

    def test_add_trunk_vlan_on_interface_that_already_has_it_does_nothing(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>PATATE</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>trunk</interface-mode>
                      <vlan-id-list>900-1100</vlan-id-list>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").never()

        self.switch.add_trunk_vlan("xe-0/0/6", 1000)

    def test_remove_ip_from_vlan(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                      <unit>
                        <name>1234</name>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>irb</name>
                  <unit>
                    <name>1234</name>
                    <family>
                      <inet>
                        <address>
                            <name>3.3.3.2/27</name>
                        </address>
                      </inet>
                    </family>
                  </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                      <unit>
                        <name>1234</name>
                        <family>
                          <inet>
                            <address operation="delete">
                              <name>3.3.3.2/27</name>
                            </address>
                          </inet>
                        </family>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>""")).and_return(an_ok_response())

        self.switch.remove_ip_from_vlan(vlan_number=1234, ip_network=IPNetwork("3.3.3.2/27"))

    def test_remove_ip_from_vlan_ip_not_found(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                      <unit>
                        <name>1234</name>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>irb</name>
                  <unit>
                    <name>1234</name>
                    <family>
                      <inet>
                        <address>
                            <name>4.4.4.2/27</name>
                        </address>
                      </inet>
                    </family>
                  </unit>
              </interface>
            </interfaces>
        """))

        with self.assertRaises(UnknownIP):
            self.switch.remove_ip_from_vlan(vlan_number=1234, ip_network=IPNetwork("3.3.3.2/27"))

    def test_remove_ip_from_vlan_unknown_vlan_raises(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>irb</name>
                      <unit>
                        <name>1234</name>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration())

        with self.assertRaises(UnknownVlan):
            self.switch.remove_ip_from_vlan(vlan_number=1234, ip_network=IPNetwork("3.3.3.2/27"))

    def test_unset_interface_access_vlan_removes_the_vlan_id(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/1</name>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
                <interfaces>
                  <interface>
                    <name>xe-0/0/1</name>
                      <unit>
                        <name>0</name>
                        <family>
                          <bridge>
                            <interface-mode>access</interface-mode>
                            <vlan-id>999</vlan-id>
                          </bridge>
                        </family>
                      </unit>
                  </interface>
                </interfaces>
            """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/1</name>
                      <unit>
                        <name>0</name>
                        <family>
                          <bridge>
                            <vlan-id operation="delete" />
                          </bridge>
                        </family>
                      </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>""")).and_return(an_ok_response())

        self.switch.unset_interface_access_vlan("xe-0/0/1")

    def test_unset_interface_access_vlan_fails_when_not_set(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/1</name>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
                <interfaces>
                  <interface>
                    <name>xe-0/0/1</name>
                      <unit>
                        <name>0</name>
                        <family>
                          <bridge>
                            <interface-mode>access</interface-mode>
                          </bridge>
                        </family>
                      </unit>
                  </interface>
                </interfaces>
            """))

        self.netconf_mock.should_receive("edit_config").never()

        with self.assertRaises(AccessVlanNotSet) as expect:
            self.switch.unset_interface_access_vlan("xe-0/0/1")

        assert_that(str(expect.exception), equal_to("Access Vlan is not set on interface xe-0/0/1"))

    def test_unset_interface_access_vlan_unknown_interface_raises(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/1</name>
                  </interface>
                </interfaces>
              </configuration>
            </filter>
        """))   .and_return(a_configuration())

        self.netconf_mock.should_receive("edit_config").never()

        with self.assertRaises(UnknownInterface) as expect:
            self.switch.unset_interface_access_vlan("xe-0/0/1")

        assert_that(str(expect.exception), equal_to("Unknown interface xe-0/0/1"))

    def test_get_interface(self):
        self.switch.in_transaction = False
        self.netconf_mock.should_receive("get_config").with_args(source="running", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces>
                    <interface>
                        <name>xe-0/0/1</name>
                    </interface>
                </interfaces>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <interfaces>
              <interface>
                <name>xe-0/0/1</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
            <bridge-domains/>
        """))

        interface = self.switch.get_interface('xe-0/0/1')

        assert_that(interface.name, equal_to("xe-0/0/1"))
        assert_that(interface.shutdown, equal_to(False))
        assert_that(interface.port_mode, equal_to(ACCESS))
        assert_that(interface.access_vlan, equal_to(None))
        assert_that(interface.trunk_native_vlan, equal_to(None))
        assert_that(interface.trunk_vlans, equal_to([]))
        assert_that(interface.auto_negotiation, equal_to(None))
        assert_that(interface.mtu, equal_to(None))

    def test_get_interfaces_lists_configuration_less_interfaces(self):
        self.switch.in_transaction = False

        self.netconf_mock.should_receive("rpc").with_args(is_xml("""
                    <get-interface-information>
                      <terse/>
                    </get-interface-information>
                """)).and_return(an_rpc_response(textwrap.dedent("""
                    <interface-information style="terse">
                      <physical-interface>
                        <name>
                          xe-0/0/1
                        </name>
                        <admin-status>
                          up
                        </admin-status>
                        <oper-status>
                          down
                        </oper-status>
                      </physical-interface>
                      <physical-interface>
                        <name>
                          xe-0/0/2
                        </name>
                        <admin-status>
                          down
                        </admin-status>
                        <oper-status>
                          down
                        </oper-status>
                      </physical-interface>
                    </interface-information>
                """)))

        self.netconf_mock.should_receive("get_config").with_args(source="running", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces />
                <bridge-domains />
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <interfaces />
            <bridge-domains/>
        """))

        if1, if2 = self.switch.get_interfaces()

        assert_that(if1.name, equal_to("xe-0/0/1"))
        assert_that(if1.shutdown, equal_to(False))
        assert_that(if1.port_mode, equal_to(ACCESS))
        assert_that(if1.access_vlan, equal_to(None))
        assert_that(if1.trunk_native_vlan, equal_to(None))
        assert_that(if1.trunk_vlans, equal_to([]))

        assert_that(if2.name, equal_to("xe-0/0/2"))
        assert_that(if2.shutdown, equal_to(True))

    def test_get_nonexistent_interface_raises(self):
        self.switch.in_transaction = False
        self.netconf_mock.should_receive("get_config").with_args(source="running", filter=is_xml("""
                    <filter>
                      <configuration>
                          <interfaces>
                            <interface>
                              <name>xe-0/0/INEXISTENT</name>
                            </interface>
                          </interfaces>
                        <bridge-domains/>
                      </configuration>
                    </filter>
                """)).and_return(a_configuration("""
                    <interfaces/>
                    <bridge-domains/>
                """))
        self.netconf_mock.should_receive("rpc").with_args(is_xml("""
                    <get-interface-information>
                      <terse/>
                    </get-interface-information>
                """)).and_return(an_rpc_response(textwrap.dedent("""
                    <interface-information style="terse">
                      <physical-interface>
                        <name>
                          xe-0/0/1
                        </name>
                        <admin-status>
                          down
                        </admin-status>
                        <oper-status>
                          down
                        </oper-status>
                      </physical-interface>
                    </interface-information>
                """)))

        with self.assertRaises(UnknownInterface) as expect:
            self.switch.get_interface('xe-0/0/INEXISTENT')

        assert_that(str(expect.exception), equal_to("Unknown interface xe-0/0/INEXISTENT"))

    def test_get_unconfigured_interface_could_be_disabled(self):
        self.switch.in_transaction = False
        self.netconf_mock.should_receive("get_config").with_args(source="running", filter=is_xml("""
                        <filter>
                          <configuration>
                              <interfaces>
                                <interface>
                                  <name>xe-0/0/27</name>
                                </interface>
                              </interfaces>
                            <bridge-domains/>
                          </configuration>
                        </filter>
                    """)).and_return(a_configuration("""
                        <interfaces/>
                        <bridge-domains/>
                    """))
        self.netconf_mock.should_receive("rpc").with_args(is_xml("""
                        <get-interface-information>
                          <terse/>
                        </get-interface-information>
                    """)).and_return(an_rpc_response(textwrap.dedent("""
                        <interface-information style="terse">
                          <physical-interface>
                            <name>
                              xe-0/0/27
                            </name>
                            <admin-status>
                              down
                            </admin-status>
                            <oper-status>
                              down
                            </oper-status>
                          </physical-interface>
                        </interface-information>
                    """)))

        assert_that(self.switch.get_interface('xe-0/0/27').shutdown, equal_to(True))

    def test_remove_trunk_vlan_removes_the_vlan_lists_in_every_possible_way(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>VLAN_NAME</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>trunk</interface-mode>
                        <vlan-id-list>1000</vlan-id-list>
                        <vlan-id-list>1000-1001</vlan-id-list>
                        <vlan-id-list>999-1000</vlan-id-list>
                        <vlan-id-list>999-1001</vlan-id-list>
                        <vlan-id-list>998-1002</vlan-id-list>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/6</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                            <vlan-id-list operation="delete">1000</vlan-id-list>
                            <vlan-id-list operation="delete">1000-1001</vlan-id-list>
                            <vlan-id-list>1001</vlan-id-list>
                            <vlan-id-list operation="delete">999-1000</vlan-id-list>
                            <vlan-id-list>999</vlan-id-list>
                            <vlan-id-list operation="delete">999-1001</vlan-id-list>
                            <vlan-id-list>999</vlan-id-list>
                            <vlan-id-list>1001</vlan-id-list>
                            <vlan-id-list operation="delete">998-1002</vlan-id-list>
                            <vlan-id-list>998-999</vlan-id-list>
                            <vlan-id-list>1001-1002</vlan-id-list>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.remove_trunk_vlan("xe-0/0/6", 1000)

    def test_remove_trunk_vlan_removes_the_vlan_even_if_referenced_by_name(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>VLAN_NAME</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>trunk</interface-mode>
                        <vlan-id-list>1000</vlan-id-list>
                        <vlan-id-list>VLAN_NAME</vlan-id-list>
                        <vlan-id-list>SOEMTHING</vlan-id-list>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").once().with_args(target="candidate", config=is_xml("""
            <config>
              <configuration>
                <interfaces>
                  <interface>
                    <name>xe-0/0/6</name>
                    <unit>
                      <name>0</name>
                      <family>
                        <bridge>
                            <vlan-id-list operation="delete">1000</vlan-id-list>
                            <vlan-id-list operation="delete">VLAN_NAME</vlan-id-list>
                        </bridge>
                      </family>
                    </unit>
                  </interface>
                </interfaces>
              </configuration>
            </config>
        """)).and_return(an_ok_response())

        self.switch.remove_trunk_vlan("xe-0/0/6", 1000)

    def test_remove_trunk_vlan_not_in_lists_raises(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>VLAN_NAME</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>trunk</interface-mode>
                        <vlan-id-list>500-999</vlan-id-list>
                        <vlan-id-list>1001-4000</vlan-id-list>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").never()

        with self.assertRaises(TrunkVlanNotSet) as expect:
            self.switch.remove_trunk_vlan("xe-0/0/6", 1000)

        assert_that(str(expect.exception), contains_string("Trunk Vlan is not set on interface xe-0/0/6"))

    def test_remove_trunk_vlan_on_access_with_the_correct_vlan_interface_raises(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>VLAN_NAME</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <interface-mode>access</interface-mode>
                      <vlan-id-list>1000</vlan-id-list>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").never()

        with self.assertRaises(InterfaceInWrongPortMode) as expect:
            self.switch.remove_trunk_vlan("xe-0/0/6", 1000)

        assert_that(str(expect.exception), contains_string("Operation cannot be performed on a access mode interface"))

    def test_remove_trunk_vlan_on_no_port_mode_interface_with_the_correct_vlan_raises(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>VLAN_NAME</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
            <interfaces>
              <interface>
                <name>xe-0/0/6</name>
                <unit>
                  <name>0</name>
                  <family>
                    <bridge>
                      <vlan-id>1000</vlan-id>
                    </bridge>
                  </family>
                </unit>
              </interface>
            </interfaces>
        """))

        self.netconf_mock.should_receive("edit_config").never()

        with self.assertRaises(InterfaceInWrongPortMode) as expect:
            self.switch.remove_trunk_vlan("xe-0/0/6", 1000)

        assert_that(str(expect.exception), contains_string("Operation cannot be performed on a access mode interface"))

    def test_remove_trunk_vlan_on_unknown_interface_raises(self):
        self.netconf_mock.should_receive("get_config").with_args(source="candidate", filter=is_xml("""
            <filter>
              <configuration>
                <interfaces/>
                <bridge-domains/>
              </configuration>
            </filter>
        """)).and_return(a_configuration("""
            <bridge-domains>
              <domain>
                <name>VLAN_NAME</name>
                <vlan-id>1000</vlan-id>
              </domain>
            </bridge-domains>
        """))

        self.netconf_mock.should_receive("edit_config").never()

        with self.assertRaises(UnknownInterface) as expect:
            self.switch.remove_trunk_vlan("xe-0/0/6", 1000)

        assert_that(str(expect.exception), contains_string("Unknown interface xe-0/0/6"))



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flexmock import flexmock, flexmock_teardown
from hamcrest import assert_that, equal_to

from netman.api.api_utils import RegexConverter
from netman.api.switch_api import SwitchApi
from netman.api.switch_session_api import SwitchSessionApi
from netman.core.objects.exceptions import UnknownSession
from netman.core.objects.unicast_rpf_modes import STRICT
from tests.api.base_api_test import BaseApiTest


class SwitchApiUnicastTest(BaseApiTest):
    def setUp(self):
        super(SwitchApiUnicastTest, self).setUp()
        self.app.url_map.converters['regex'] = RegexConverter

        self.switch_factory = flexmock()
        self.switch_mock = flexmock()
        self.session_manager = flexmock()

        self.session_manager.should_receive("get_switch_for_session").and_raise(UnknownSession("patate"))

        SwitchApi(self.switch_factory, self.session_manager).hook_to(self.app)
        SwitchSessionApi(self.switch_factory, self.session_manager).hook_to(self.app)

    def tearDown(self):
        flexmock_teardown()

    def test_set_unicast_rpf_mode_strict(self):
        self.switch_factory.should_receive('get_switch').with_args('my.switch').and_return(self.switch_mock).once().ordered()
        self.switch_mock.should_receive('connect').once().ordered()
        self.switch_mock.should_receive('set_vlan_unicast_rpf_mode').with_args(2500, str(STRICT)).once().ordered()
        self.switch_mock.should_receive('disconnect').once().ordered()

        result, code = self.put("/switches/my.switch/vlans/2500/unicast-rpf-mode", raw_data=STRICT)

        assert_that(code, equal_to(204))

    def test_set_unicast_rpf_mode_invalid(self):
        result, code = self.put("/switches/my.switch/vlans/2500/unicast-rpf-mode", raw_data='shizzle')

        assert_that(code, equal_to(400))
        assert_that(result, equal_to({'error': 'Invalid unicast rpf mode'}))

    def test_unset_unicast_rpf_mode(self):
        self.switch_factory.should_receive('get_switch').with_args('my.switch').and_return(self.switch_mock).once().ordered()
        self.switch_mock.should_receive('connect').once().ordered()
        self.switch_mock.should_receive('unset_vlan_unicast_rpf_mode').with_args(2500).once().ordered()
        self.switch_mock.should_receive('disconnect').once().ordered()

        result, code = self.delete("/switches/my.switch/vlans/2500/unicast-rpf-mode")

        assert_that(code, equal_to(204))



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hamcrest import assert_that, is_, none, is_not
from netaddr import IPAddress

from tests.adapters.configured_test_case import ConfiguredTestCase, skip_on_switches


class DhcpRelayServerTest(ConfiguredTestCase):
    __test__ = False

    @skip_on_switches("juniper", "juniper_qfx_copper", "juniper_mx", "dell", "dell_telnet", "dell10g", "dell10g_telnet")
    def test_add_and_get_and_delete_dhcp_relay_server(self):
        try:
            self.client.add_vlan(2999, name="my-test-vlan")
            self.client.add_dhcp_relay_server(2999, ip_address=IPAddress("10.10.10.1"))

            vlan = self.get_vlan_from_list(2999)
            dhcp_relay_server = next((g for g in vlan.dhcp_relay_servers if str(g) == '10.10.10.1'), None)

            assert_that(dhcp_relay_server, is_not(None))

            self.client.remove_dhcp_relay_server(2999, ip_address=IPAddress("10.10.10.1"))

            vlan = self.get_vlan_from_list(2999)
            dhcp_relay_server = next((g for g in vlan.dhcp_relay_servers if str(g) == '10.10.10.1'), None)
            assert_that(dhcp_relay_server, is_(none()))

        finally:
            self.client.remove_vlan(2999)



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hamcrest import assert_that, contains_inanyorder

from netman.core.objects.exceptions import UnknownVlan
from tests import has_message
from tests.adapters.compliance_test_case import ComplianceTestCase


class GetVlanInterfacesTest(ComplianceTestCase):
    _dev_sample = "cisco"

    def setUp(self):
        super(GetVlanInterfacesTest, self).setUp()

    def test_returns_vlan_interfaces(self):
        self.client.add_vlan(1000, name="vlan1000")

        self.try_to.set_access_mode(self.test_ports[0].name)
        self.try_to.set_access_vlan(self.test_ports[0].name, 1000)

        self.try_to.set_trunk_mode(self.test_ports[1].name)
        self.try_to.add_trunk_vlan(self.test_ports[1].name, 1000)

        # At this time, interface_native_vlan is not supported on get_vlan_interfaces

        assert_that(self.client.get_vlan_interfaces(1000),
                    contains_inanyorder(self.test_ports[0].name, self.test_ports[1].name))

    def test_fails_when_the_vlan_does_not_exist(self):
        with self.assertRaises(UnknownVlan)as expect:
            self.client.get_vlan_interfaces(2000)

        assert_that(expect.exception, has_message("Vlan 2000 not found"))

    def tearDown(self):
        self.janitor.unset_interface_access_vlan(self.test_ports[0].name)

        self.janitor.remove_trunk_vlan(self.test_ports[1].name, 1000)
        self.janitor.set_access_mode(self.test_ports[1].name)

        self.janitor.remove_vlan(1000)

        super(GetVlanInterfacesTest, self).tearDown()



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings

from netman.core.objects.interface_states import OFF, ON


class BackwardCompatibleSwitchOperations(object):
    """
    Depecrated methods
    """
    def remove_access_vlan(self, interface_id):
        warnings.warn("Deprecated, use unset_interface_access_vlan(interface_id) instead", DeprecationWarning)
        return self.unset_interface_access_vlan(interface_id)

    def configure_native_vlan(self, interface_id, vlan):
        warnings.warn("Deprecated, use set_interface_native_vlan(interface_id, vlan) instead", DeprecationWarning)
        return self.set_interface_native_vlan(interface_id, vlan)

    def remove_native_vlan(self, interface_id):
        warnings.warn("Deprecated, use unset_interface_native_vlan(interface_id) instead", DeprecationWarning)
        return self.unset_interface_native_vlan(interface_id)

    def remove_vlan_access_group(self, vlan_number, direction):
        warnings.warn("Deprecated, use unset_vlan_access_group(vlan_number, direction) instead", DeprecationWarning)
        return self.unset_vlan_access_group(vlan_number, direction)

    def remove_vlan_vrf(self, vlan_number):
        warnings.warn("Deprecated, use unset_vlan_vrf(vlan_number) instead", DeprecationWarning)
        return self.unset_vlan_vrf(vlan_number)

    def remove_interface_description(self, interface_id):
        warnings.warn("Deprecated, use unset_interface_description(interface_id) instead", DeprecationWarning)
        return self.unset_interface_description(interface_id)

    def remove_bond_description(self, number):
        warnings.warn("Deprecated, use unset_bond_description(number) instead", DeprecationWarning)
        return self.unset_bond_description(number)

    def configure_bond_native_vlan(self, number, vlan):
        warnings.warn("Deprecated, use set_bond_native_vlan(number, vlan) instead", DeprecationWarning)
        return self.set_bond_native_vlan(number, vlan)

    def remove_bond_native_vlan(self, number):
        warnings.warn("Deprecated, use unset_bond_native_vlan(number) instead", DeprecationWarning)
        return self.unset_bond_native_vlan(number)

    def enable_lldp(self, interface_id, enabled):
        warnings.warn("Deprecated, use set_interface_lldp_state(interface_id, enabled) instead", DeprecationWarning)
        return self.set_interface_lldp_state(interface_id, enabled)

    def shutdown_interface(self, interface_id):
        warnings.warn("Deprecated, use set_interface_state(interface_id, state) instead", DeprecationWarning)
        return self.set_interface_state(interface_id, OFF)

    def openup_interface(self, interface_id):
        warnings.warn("Deprecated, use set_interface_state(interface_id, state) instead", DeprecationWarning)
        return self.set_interface_state(interface_id, ON)



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netman.api.objects import sub_dict, Serializer, Serializers
from netman.api.objects import base_interface
from netman.core.objects.interface import Interface


__all__ = ['to_api', 'to_core']


class V1(Serializer):
    since_version = 1

    def to_api(self, interface):
        return dict(
            base_interface.to_api(interface),
            name=interface.name,
            bond_master=interface.bond_master,
            auto_negotiation=interface.auto_negotiation
        )

    def to_core(self, serialized):
        params = dict(vars(base_interface.to_core(serialized)))
        params.update(sub_dict(serialized, 'name', 'bond_master', 'auto_negotiation'))
        return Interface(**params)


serializers = Serializers(V1())

to_api = serializers.to_api
to_core = serializers.to_core



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hamcrest import assert_that, is_
from netaddr import IPNetwork, IPAddress

from netman.core.objects.access_groups import IN, OUT
from netman.core.objects.exceptions import UnknownVlan
from tests import has_message
from tests.adapters.compliance_test_case import ComplianceTestCase


class GetVlanTest(ComplianceTestCase):
    _dev_sample = "cisco"

    def test_returns_a_vlan_with_all_available_data(self):
        self.client.add_vlan(1000, name="vlan_name")

        self.try_to.set_vlan_vrf(1000, "DEFAULT-LAN")
        self.try_to.add_ip_to_vlan(1000, IPNetwork("10.10.10.2/29"))
        self.try_to.set_vlan_access_group(1000, IN, "ACL-IN")
        self.try_to.set_vlan_access_group(1000, OUT, "ACL-OUT")
        self.try_to.add_vrrp_group(1000, 1,
                                   ips=[IPAddress("10.10.10.1")], priority=110, hello_interval=10,
                                   dead_interval=11, track_id=self.test_vrrp_track_id, track_decrement=50)
        self.try_to.add_dhcp_relay_server(1000, IPAddress("11.11.11.1"))
        self.try_to.set_vlan_icmp_redirects_state(1000, False)

        assert_that(self.client.get_vlan(1000), is_(self.get_vlan_from_list(1000)))

    def test_fails_when_the_vlan_does_not_exist(self):
        with self.assertRaises(UnknownVlan) as expect:
            self.client.get_vlan(2000)

        assert_that(expect.exception, has_message("Vlan 2000 not found"))

    def tearDown(self):
        self.janitor.remove_vlan(1000)
        super(GetVlanTest, self).tearDown()



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from hamcrest import assert_that, instance_of, is_, is_not
import mock
from netman.core.objects.flow_control_switch import FlowControlSwitch

from netman.core import switch_factory

from netman.core.objects.switch_base import SwitchBase
from netman.adapters.switches.remote import RemoteSwitch
from netman.core.objects.switch_descriptor import SwitchDescriptor
from netman.core.switch_factory import SwitchFactory


class SwitchFactoryTest(unittest.TestCase):

    def setUp(self):
        self.semaphore_mocks = {}
        self.factory = SwitchFactory(switch_source=None, lock_factory=MockLockFactory(self.semaphore_mocks))

        switch_factory.factories['test_model'] = _FakeSwitch

    def tearDown(self):
        switch_factory.factories.pop('test_model')

    def test_get_connection_to_anonymous_switch(self):
        my_semaphore = mock.Mock()
        self.semaphore_mocks['hostname'] = my_semaphore

        switch = self.factory.get_anonymous_switch(hostname='hostname', model='test_model', username='username',
                                                   password='password', port=22)

        assert_that(switch, is_(instance_of(FlowControlSwitch)))
        assert_that(switch.wrapped_switch, is_(instance_of(_FakeSwitch)))
        assert_that(switch.lock, is_(my_semaphore))
        assert_that(switch.wrapped_switch.switch_descriptor, is_(
                SwitchDescriptor(hostname='hostname', model='test_model', username='username',
                                 password='password', port=22)))

    def test_two_get_connections_on_the_same_switch_should_give_the_same_semaphore(self):
        self.semaphore_mocks['hostname'] = mock.Mock()

        switch1 = self.factory.get_anonymous_switch(hostname='hostname', model='test_model')
        switch2 = self.factory.get_anonymous_switch(hostname='hostname', model='test_model')

        assert_that(switch1.lock, is_(switch2.lock))

    def test_two_get_connections_on_different_switches_should_give_different_semaphores(self):
        self.semaphore_mocks['hostname1'] = mock.Mock()
        self.semaphore_mocks['hostname2'] = mock.Mock()

        switch1 = self.factory.get_anonymous_switch(hostname='hostname1', model='test_model')
        switch2 = self.factory.get_anonymous_switch(hostname='hostname2', model='test_model')

        assert_that(switch1.lock, is_not(switch2.lock))

    def test_get_connection_to_anonymous_remote_switch(self):
        my_semaphore = mock.Mock()
        self.semaphore_mocks['hostname'] = my_semaphore
        switch = self.factory.get_anonymous_switch(hostname='hostname', model='test_model', username='username',
                                                   password='password', port=22,
                                                   netman_server='https://netman.url.example.org:4443')

        assert_that(switch, is_(instance_of(FlowControlSwitch)))
        assert_that(switch.wrapped_switch, is_(instance_of(RemoteSwitch)))
        assert_that(switch.lock, is_(my_semaphore))
        assert_that(switch.wrapped_switch.switch_descriptor, is_(
                SwitchDescriptor(hostname='hostname', model='test_model', username='username',
                                 password='password', port=22,
                                 netman_server='https://netman.url.example.org:4443')))

    def test_get_switch_by_descriptor(self):
        my_semaphore = mock.Mock()
        self.semaphore_mocks['hostname'] = my_semaphore

        switch = self.factory.get_switch_by_descriptor(SwitchDescriptor(model='test_model', hostname='hostname'))

        assert_that(switch, is_(instance_of(FlowControlSwitch)))
        assert_that(switch.wrapped_switch, is_(instance_of(_FakeSwitch)))
        assert_that(switch.lock, is_(my_semaphore))
        assert_that(switch.wrapped_switch.switch_descriptor,
                    is_(SwitchDescriptor(model='test_model', hostname='hostname')))


class MockLockFactory(object):

    def __init__(self, mock_dict):
        self.mock_dict = mock_dict

    def new_lock(self, name, timeout=0):
        return self.mock_dict.pop(name)


class _FakeSwitch(SwitchBase):
    pass



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netman.core.objects.access_groups import IN, OUT


class NetmanException(Exception):
    pass


class InvalidValue(NetmanException):
    def __init__(self, msg="Invalid Value"):
        super(InvalidValue, self).__init__(msg)


class UnknownResource(NetmanException):
    def __init__(self, msg="Resource not found"):
        super(UnknownResource, self).__init__(msg)


class Conflict(NetmanException):
    def __init__(self, msg="Conflicting value"):
        super(Conflict, self).__init__(msg)


class SessionAlreadyExists(Conflict):
    def __init__(self, session_id=None):
        super(SessionAlreadyExists, self).__init__(msg="Session ID already exists: {}".format(session_id))


class UnavailableResource(NetmanException):
    def __init__(self, msg="Resource not available"):
        super(UnavailableResource, self).__init__(msg)


class OperationNotCompleted(NetmanException):
    def __init__(self, problem=None):
        super(OperationNotCompleted, self).__init__("An error occured while completing operation, no modifications have been applied : {0}".format(problem))


class InterfaceResetIncomplete(NetmanException):
    def __init__(self, interface_data=None):
        super(InterfaceResetIncomplete, self).__init__("The interface reset has failed to remove these properties: {0}".format(interface_data))


class UnknownVlan(UnknownResource):
    def __init__(self, vlan_number=None):
        super(UnknownVlan, self).__init__("Vlan {} not found".format(vlan_number))


class UnknownInterface(UnknownResource):
    def __init__(self, interface=None):
        super(UnknownInterface, self).__init__("Unknown interface {}".format(interface))


class UnknownIP(UnknownResource):
    def __init__(self, ip_network=None):
        super(UnknownIP, self).__init__("IP {} not found".format(ip_network))


class UnknownAccessGroup(UnknownResource):
    def __init__(self, direction=None):
        super(UnknownAccessGroup, self).__init__("{} IP access group not found".format({IN: "Inbound", OUT: "Outgoing"}[direction] if direction else ""))


class UnknownSession(UnknownResource):
    def __init__(self, session_id=None):
        super(UnknownSession, self).__init__("Session \"{}\" not found.".format(session_id))


class UnknownVrf(UnknownResource):
    def __init__(self, name=None):
        super(UnknownVrf, self).__init__("VRF name \"{}\" was not configured.".format(name))


class UnknownDhcpRelayServer(UnknownResource):
    def __init__(self, vlan_number=None, ip_address=None):
        super(UnknownDhcpRelayServer, self).__init__("DHCP relay server {} not found on VLAN {}".format(ip_address, vlan_number))


class DhcpRelayServerAlreadyExists(UnknownResource):
    def __init__(self, vlan_number=None, ip_address=None):
        super(DhcpRelayServerAlreadyExists, self).__init__("DHCP relay server {} already exists on VLAN {}".format(ip_address, vlan_number))


class AccessVlanNotSet(UnknownResource):
    def __init__(self, interface=None):
        super(AccessVlanNotSet, self).__init__("Access Vlan is not set on interface {}".format(interface))


class TrunkVlanNotSet(UnknownResource):
    def __init__(self, interface=None):
        super(TrunkVlanNotSet, self).__init__("Trunk Vlan is not set on interface {}".format(interface))


class NativeVlanNotSet(UnknownResource):
    def __init__(self, interface=None):
        super(NativeVlanNotSet, self).__init__("Trunk native Vlan is not set on interface {}".format(interface))


class InterfaceSpanningTreeNotEnabled(UnknownResource):
    def __init__(self, interface=None):
        super(InterfaceSpanningTreeNotEnabled, self).__init__("Spanning tree is not enabled on interface {}".format(interface))


class VlanVrfNotSet(UnknownResource):
    def __init__(self, vlan=None):
        super(VlanVrfNotSet, self).__init__("VRF is not set on vlan {}".format(vlan))


class IPNotAvailable(Conflict):
    def __init__(self, ip_network=None, reason=None):
        super(IPNotAvailable, self).__init__("IP {} is not available in this vlan{}".format(ip_network, (": " + reason) if reason is not None else ""))


class IPAlreadySet(Conflict):
    def __init__(self, ip_network=None, present_ip_network=None):
        super(IPAlreadySet, self).__init__("IP {} is already present in this vlan as {}".format(ip_network, present_ip_network))


class VlanAlreadyExist(Conflict):
    def __init__(self, vlan_number=None):
        super(VlanAlreadyExist, self).__init__("Vlan {} already exists".format(vlan_number))


class InterfaceInWrongPortMode(Conflict):
    def __init__(self, mode=None):
        super(InterfaceInWrongPortMode, self).__init__("Operation cannot be performed on a {} mode interface".format(mode))


class VlanAlreadyInTrunk(Conflict):
    def __init__(self, vlan=None):
        super(VlanAlreadyInTrunk, self).__init__("Vlan {} cannot be set as native vlan because it is already a member of the trunk".format(vlan))


class VrrpAlreadyExistsForVlan(Conflict):
    def __init__(self, vlan=None, vrrp_group_id=None):
        super(VrrpAlreadyExistsForVlan, self).__init__("Vrrp group {group} is already in use on vlan {vlan}".format(group=vrrp_group_id, vlan=vlan))


class VrrpDoesNotExistForVlan(InvalidValue):
    def __init__(self, vlan=None, vrrp_group_id=None):
        super(VrrpDoesNotExistForVlan, self).__init__("Vrrp group {group} does not exist for vlan {vlan}".format(group=vrrp_group_id, vlan=vlan))


class VarpAlreadyExistsForVlan(Conflict):
    def __init__(self, vlan=None, ip_network=None):
        super(VarpAlreadyExistsForVlan, self).__init__("Varp ip network {ip} is already on vlan {vlan}".format(ip=ip_network, vlan=vlan))


class VarpDoesNotExistForVlan(InvalidValue):
    def __init__(self, vlan=None, ip_network=None):
        super(VarpDoesNotExistForVlan, self).__init__("Varp ip {ip} is not configured for vlan {vlan}".format(ip=ip_network, vlan=vlan))


class NoIpOnVlanForVrrp(InvalidValue):
    def __init__(self, vlan=None):
        super(NoIpOnVlanForVrrp, self).__init__("Vlan {vlan} needs an IP before configuring VRRP".format(vlan=vlan))


class BadVlanNumber(InvalidValue):
    def __init__(self):
        super(BadVlanNumber, self).__init__("Vlan number is invalid")


class BadInterfaceDescription(InvalidValue):
    def __init__(self, desc=None):
        super(BadInterfaceDescription, self).__init__("Invalid description : {}".format(desc))


class BadVrrpGroupNumber(InvalidValue):
    def __init__(self, minimum=None, maximum=None):
        super(BadVrrpGroupNumber, self).__init__("VRRP group number is invalid, must be contained between {min} and {max}".format(min=minimum, max=maximum))


class BadVrrpPriorityNumber(InvalidValue):
    def __init__(self, minimum=None, maximum=None):
        super(BadVrrpPriorityNumber, self).__init__("VRRP priority value is invalid, must be contained between {min} and {max}".format(min=minimum, max=maximum))


class BadVrrpTimers(InvalidValue):
    def __init__(self):
        super(BadVrrpTimers, self).__init__("VRRP timers values are invalid")


class BadVrrpAuthentication(InvalidValue):
    def __init__(self):
        super(BadVrrpAuthentication, self).__init__("VRRP authentication is invalid")


class BadVrrpTracking(InvalidValue):
    def __init__(self):
        super(BadVrrpTracking, self).__init__("VRRP tracking values are invalid")


class BadVlanName(InvalidValue):
    def __init__(self):
        super(BadVlanName, self).__init__("Vlan name is invalid")


class LockedSwitch(UnavailableResource):
    def __init__(self):
        super(LockedSwitch, self).__init__("Switch is locked and can't be modified")


class UnableToAcquireLock(UnavailableResource):
    def __init__(self):
        super(UnableToAcquireLock, self).__init__("Unable to acquire a lock in a timely fashion")


class BadBondNumber(InvalidValue):
    def __init__(self):
        super(BadBondNumber, self).__init__("Bond number is invalid")


class InterfaceNotInBond(UnknownResource):
    def __init__(self):
        super(InterfaceNotInBond, self).__init__("Interface not associated to specified bond")


class BondAlreadyExist(Conflict):
    def __init__(self, number=None):
        super(BondAlreadyExist, self).__init__("Bond {} already exists".format(number))


class UnknownBond(UnknownResource):
    def __init__(self, number=None):
        super(UnknownBond, self).__init__("Bond {} not found".format(number))


class BadBondLinkSpeed(InvalidValue):
    def __init__(self):
        super(BadBondLinkSpeed, self).__init__("Malformed bond link speed")


class UnknownSwitch(UnknownResource):
    def __init__(self, name=None):
        super(UnknownSwitch, self).__init__("Switch \"{0}\" is not configured".format(name))


class MalformedSwitchSessionRequest(InvalidValue):
    def __init__(self):
        super(MalformedSwitchSessionRequest, self).__init__("Malformed switch session request")


class BadLoadIntervalNumber(InvalidValue):
    def __init__(self):
        super(BadLoadIntervalNumber, self).__init__("Load interval number is invalid")


class BadMplsIpState(InvalidValue):
    def __init__(self, state=None):
        super(BadMplsIpState, self).__init__("MPLS IP state is invalid : {}".format(state))


class Timeout(Exception):
    pass


class ConnectTimeout(Exception):
    def __init__(self, host=None, port=None):
        super(ConnectTimeout, self).__init__("Timed out while connecting to {} on port {}".format(host, port))


class CommandTimeout(Exception):
    def __init__(self, wait_for=None, buffer=None):
        super(CommandTimeout, self).__init__("Command timed out expecting {}. Current read buffer: {}"
                                             .format(repr(wait_for), buffer))


class PrivilegedAccessRefused(Exception):
    def __init__(self, buffer=None):
        super(PrivilegedAccessRefused, self).__init__("Could not get PRIVILEGED exec mode. "
                                                      "Current read buffer: {}".
                                                      format(buffer))


class CouldNotConnect(Exception):
    def __init__(self, host=None, port=None):
        super(CouldNotConnect, self).__init__("Could not connect to {} on port {}".format(host, port))


class InvalidAccessGroupName(InvalidValue):
    def __init__(self, name=None):
        super(InvalidAccessGroupName, self).__init__("Access Group Name is invalid: {}".format(name))


class InvalidMtuSize(InvalidValue):
    def __init__(self, err_msg=None):
        super(InvalidMtuSize, self).__init__("MTU value is invalid : {}".format(err_msg))


class InvalidUnicastRPFMode(InvalidValue):
    def __init__(self, mode=None):
        super(InvalidUnicastRPFMode, self).__init__("Unknown Unicast RPF mode: \"{}\"".format(mode))


class UnsupportedOperation(NotImplementedError):
    def __init__(self, operation=None, message=None):
        super(UnsupportedOperation, self).__init__("Operation \"{}\" is not supported on this equipment: {}".format(operation, message))



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps
import json
import logging

from flask import make_response, request, Response, current_app
from werkzeug.routing import BaseConverter
from netman.api import NETMAN_API_VERSION

from netman.core.objects.exceptions import UnknownResource, Conflict, InvalidValue


def to_response(fn):
    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        try:
            result = fn(self, *args, **kwargs)
            if isinstance(result, Response):
                return result
            else:
                code, data = result
                if data is not None:
                    response = json_response(data, code)
                else:
                    response = make_response("", code)
        except InvalidValue as e:
            response = exception_to_response(e, 400)
        except UnknownResource as e:
            response = exception_to_response(e, 404)
        except Conflict as e:
            response = exception_to_response(e, 409)
        except NotImplementedError as e:
            response = exception_to_response(e, 501)
        except Exception as e:
            logging.exception(e)
            response = exception_to_response(e, 500)

        self.logger.info("Responding {} : {}".format(response.status_code, response.data))
        if 'Netman-Max-Version' in request.headers:
            response.headers['Netman-Version'] = min(
                float(request.headers['Netman-Max-Version']),
                NETMAN_API_VERSION)
        return response

    return wrapper


def exception_to_response(exception, code):
    data = {'error': str(exception)}

    if "Netman-Verbose-Errors" in request.headers:
        if hasattr(exception, "__module__"):
            data["error-module"] = exception.__module__
        data["error-class"] = exception.__class__.__name__
    else:
        if data['error'] == "":
            if hasattr(exception, "__module__"):
                data['error'] = "Unexpected error: {}.{}".format(exception.__module__, exception.__class__.__name__)
            else:
                data['error'] = "Unexpected error: {}".format(exception.__class__.__name__)

    response = json_response(data, code)
    response.status_code = code

    return response


def json_response(data, code):

    json_data = json.dumps(data, indent=None)
    response = current_app.response_class(json_data, mimetype='application/json; charset=UTF-8')
    response.status_code = code

    return response


class RegexConverter(BaseConverter):
    def __init__(self, url_map, *items):
        super(RegexConverter, self).__init__(url_map)
        self.regex = items[0]


class BadRequest(InvalidValue):
    pass


class MultiContext(object):
    def __init__(self, switch_api, parameters,  *contexts):
        self.context_instances = []
        for context in contexts:
            obj = context(switch_api)
            obj.process(parameters)
            self.context_instances.append(obj)

        self.parameters = parameters

    def __enter__(self):
        return [(obj.__enter__()) for obj in self.context_instances]

    def __exit__(self, type_, value, traceback):
        for context in self.context_instances:
            context.__exit__(type_, value, traceback)
